/* irsender.cc       letzte Aenderung: 2.7.2012
 Version: 0.1

 Prozessor: ATmega8A mit 16MHz Quarz
 Schaltung: Port B0 mit Frequenzmesser oder Oszilloskop verbunden
                 B1 mit LED
                 B2 mit IR-LED
            Tastenmatrix
		 D2 (INT0) Matrix-Ausgaenge ueber Dioden
		 D3 (INT1) nicht benutzt
		 fuer 4*5-Matrix:
		 C0-C5 Matrix-Eingaenge, 6 Reihen, als Schutz ueber 330-Ohm-Widerstaende anschliessen
		 D4-D7 Matrix-Ausgaenge, 4 Spalten

Skizze unterste 2 Reihen:
        /                /                /                /
C1 ---./ ,---+   C1 ---./ ,---+   C1 ---./ ,---+   C1 ---./ ,---+
             |                |                |                |
        /    |           /    |           /    |           /    |
C0 ---./ .---+   C0 ---./ .---+   C0 ---./ .---+   C0 ---./ .---+
             |                |                |                |
             |                |                |                |
            D4               D5               D6               D7
            _|_              _|_              _|_              _|_
            / \              / \              / \              / \
           /___\            /___\            /___\            /___\
             |                |                |                |
             +----------------+----------------+----------------+---INT0

 Autor: Rolf Pfister
 Copyright: Freeware
 History:
 Datum     Vers Kommentar
 1.7.2012  0.0  Erstellung
 2.7.12    0.1  mit Tastenmatrix

*/
#include <avr/io.h>
#include <avr/interrupt.h>
#include "ulong.h"

#define TAKT_BIT 0  //zur Ausgabe der Modulationsfrequenz fuer Testzwecke 
#define LED1_BIT 1  //erste Kontroll-LED
#define LED2_BIT 3  //zweite Kontroll-LED
#define IRLED_BIT 2 //Ausgabe des Sendesignals

#define INT0_BIT PD2

#define NEC_FORMAT  //Format fuer Kaon
#ifdef NEC_FORMAT
#define F_TAKT 38  //Modulations-Frequenz in kHz
#define PULSBREITE 21 //Anzahl Takte pro Puls
#define NULLBITLUECKE 21 //Anzahl Takte fuer 0-Bit
#define EINSBITLUECKE 63 //Anzahl Takte fuer 1-Bit
#define STARTHI 336 //Anzahl Takte fuer Startbit high
#define STARTLO 158 //Anzahl Takte fuer Startbit low
#define STARTLOREP 84 //Anzahl Takte fuer Startbit low im Repeat-Block
#define ADRESSE 0x0410 //Adresse des Geraetes
#define COD_POWER 0x00 //Code fuer Power-Taste
#define COD_1     0x38 //Code fuer 1-Taste
#define COD_2     0xB8 //Code fuer 2-Taste
#define COD_3     0x78 //Code fuer 3-Taste
#define COD_4     0x02 //Code fuer 4-Taste
#define COD_5     0x82 //Code fuer 5-Taste
#define COD_6     0x42 //Code fuer 6-Taste
#define COD_7     0x22 //Code fuer 7-Taste
#define COD_8     0xA2 //Code fuer 8-Taste
#define COD_9     0x62 //Code fuer 9-Taste
#define COD_0     0xE2 //Code fuer 0-Taste
#define COD_NEXT  0x60 //Code fuer Sender+
#define COD_PREV  0xE0 //Code fuer Sender-
#define COD_VPLUS 0x40 //Code fuer VOL+ Taste
#define COD_VMINUS 0xC0 //Code fuer VOL- Taste
#define COD_MUTE  0x20 //Code fuer Mute-Taste
#define COD_INFO  0xD8 //Code fuer Info-Taste
#define COD_EXIT  0x08 //Code fuer Exit-Taste
#define COD_RECAL 0x30 //Code fuer Recall-Taste
#endif

//provisorisch:
//#define GENAUERUNDUNG

/**** Takterzeugung ****/
static volatile uchar taktzaehler=0,sendeflag=0;
static volatile uchar int0_flag=0;

// Genaue Quarzfrequenz in Hertz:
#define F_CPU 16000000L

#ifdef GENAUERUNDUNG
#define NormalerCompwert (((F_CPU/F_TAKT+500)/2000)-1) //bei weniger 0.75 abgerundeter Wert
#define ZweiterCompwert (((F_CPU/F_TAKT+1500)/2000)-1) //bei mehr als 0.25 aufgerundeter Wert
#else
#define NormalerCompwert (((F_CPU/F_TAKT+1000)/2000)-1) //ab 0.5 aufrunden, sonst abrunden
#endif

ISR(SIG_OUTPUT_COMPARE1A)
{
 static uchar takt=0;
 if((takt^=1)!=0)
  {
   taktzaehler++;
#ifdef GENAUERUNDUNG
   OCR1A=NormalerCompwert;
#endif
   PORTB |= (1<<TAKT_BIT); //PB0 setzen
   if(sendeflag!=0) PORTB |= (1<<IRLED_BIT); //falls sendeflag aktiv: IR-LED einschalten
  }
 else
  {
#ifdef GENAUERUNDUNG
   OCR1A=ZweiterCompwert;
#endif
   PORTB &= ~(1<<TAKT_BIT); //PB0 loeschen
   PORTB &= ~(1<<IRLED_BIT); //IR-LED ausschalten
  }
}

void timer1_init()
{
 TCCR1B = (1<<WGM12)|(1<<CS10); //CTC-OCR1A-Modus, Prescaler=1 (also Timer-Takt ist gleich Quarz-Takt)
 OCR1A = NormalerCompwert;
 TCNT1 = 0; //Startwert des Timers
 TIMSK = 1<<OCIE1A; //Timer1 Interrupts bei Vergleichswert
}

//Bei einem unerwarteten Interrupt nichts machen:
void __attribute__ ((naked)) __vector_default(void) {__asm__ ("reti");}

ISR(INT0_vect)
{
 int0_flag=1;
}

void ir_startbit()
{
 taktzaehler=0;
 sendeflag=1; //Startbit starten
 PORTB |= (1<<LED2_BIT); //LED2 einschalten
 while(taktzaehler < (STARTHI/2)) ;//erste Haelfte des Startbits abwarten
 taktzaehler=0;
 while(taktzaehler < (STARTHI/2)) ;//zweite Haelfte des Startbits abwarten
 taktzaehler=0;
 sendeflag=0; //Low-Teil des Startbits starten
 PORTB &= ~(1<<LED2_BIT); //LED2 ausschalten
 while(taktzaehler < STARTLO) ;//Low-Teil des Startbits abwarten
}

void ir_bit_senden(uchar data)  //einzelnes Bit ueber IR-Diode senden
{
 uchar luecke;
 sendeflag = 1; //Puls fuer aktuelles Bit senden
 taktzaehler=0;
 PORTB |= (1<<LED2_BIT); //LED2 einschalten
 while(taktzaehler < PULSBREITE) ;//Puls fuer aktuelles Bit abwarten
 sendeflag = 0; //Luecke fuer aktuelles Bit senden
 taktzaehler=0;
 PORTB &= ~(1<<LED2_BIT); //LED2 ausschalten
 luecke = (data==0) ? NULLBITLUECKE : EINSBITLUECKE;
 while(taktzaehler < luecke) ;//Luecke fuer aktuelles Bit abwarten
}

void ir_byte_senden(uchar data)  //ein Byte ueber IR-Diode senden
{
 uchar maske;
 for(maske=0x80;maske!=0;maske>>=1)
  {
   ir_bit_senden(data&maske);
  }
}

void ir_repeat()
{
 //Hier sollte noch ein repeat-code gesendet werden,
 //ist aber fuer Kaon nicht unbedingt noetig.
 taktzaehler=0;
 sendeflag=1; //Startbit starten
 PORTB |= (1<<LED2_BIT); //LED2 einschalten
 while(taktzaehler < (STARTHI/2)) ;//erste Haelfte des Startbits abwarten
 taktzaehler=0;
 while(taktzaehler < (STARTHI/2)) ;//zweite Haelfte des Startbits abwarten
 taktzaehler=0;
 sendeflag=0; //Low-Teil des Startbits starten
 PORTB &= ~(1<<LED2_BIT); //LED2 ausschalten
 while(taktzaehler < STARTLOREP) ;//Low-Teil des Startbits abwarten
 ir_bit_senden(1);
}

void milliwait(int n)
{
 taktzaehler=0;
 while(n>0)
  {
   while(taktzaehler<F_TAKT) ;//eine Millisekunde abwarten
   taktzaehler=0;
   --n;
  }
}

static uchar tastencode[6*4] = 
 {
  COD_POWER, 0x00, 0x00, COD_MUTE,           //Power            Mute
  0x00, 0x00, 0x00, COD_VPLUS,               //                 VOL+
  COD_INFO, COD_EXIT, COD_RECAL, COD_VMINUS, //Info Exit Recall VOL-
  COD_7, COD_8, COD_9, COD_0,     //7 8 9 0
  COD_4, COD_5, COD_6, COD_NEXT,  //4 5 6 Sender+
  COD_1, COD_2, COD_3, COD_PREV   //1 2 3 Sender-
 };
int main(void)
{
 DDRB  = 0x0F; //PortB PB0-3 als Ausgaenge
 PORTB = 0; //alle Auasgaenge aus
 //DDRD = 0; //alles Eingaenge
 PORTD = (1<<PD2) | (1<<PD3) | (1<<PD4) | (1<<PD5) | (1<<PD6) | (1<<PD7); //Pullup-Widerstaende setzen
 DDRC = 0x3F; //PC0 bis PC5 als Ausgaenge
 PORTC = 0; //PC alle Ausgaenge auf LOW (=aktiv)

 //Interrupts vorbereiten:
 MCUCR |= (1<<ISC01);//INT0 bei fallender Flanke
 //MCUCR |= (1<<ISC11);//INT1 bei fallender Flanke
 GICR |= (1<<INT0);//INT0-Interrupts erlauben
 //GICR |= (1<<INT1);//INT1-Interrupts erlauben
 timer1_init();
 sei(); //Interrupt einschalten

 while(1) //Hauptschlaufe
  {
   if(int0_flag!=0)
    {
     //Entprellung:
     milliwait(10);
     if(int0_flag!=0 && (PIND & (1<<INT0_BIT))!=0) int0_flag=0; //wenn Taste schon wieder losgelassen --> Stoerpuls, ignorieren

     if(int0_flag!=0)
      {
       uchar i,j=0,n;
       for(i=0x01;i!=0x40;i<<=1) //Leitungen C0 bis C5 der Reihe nach auf Low
	{
	 PORTC = i^0xFF;
	 milliwait(1);
	 if((n=(PIND&0xF0))!=0xF0) //ist einer von PD4 bis PD7 auf low?
	  {
	   n>>=4;
	   n ^= 0x0F;
	   for(j=0x01;j!=n;j<<=1) ;//entsprechende Reihe suchen
	   break;
	  }
	}
       if(j!=0) //Reihe und Spalte der Taste gefunden
	{
	 for(n=0;j!=1;j>>=1)
	   n++; //Spalte zaehlen
	 for(;i!=1;i>>=1)
	   n+=4; //Reihe zaehlen
	 n=tastencode[n];
	 ir_startbit();
	 ir_byte_senden(ADRESSE>>8);//Geraeteadresse senden
	 ir_byte_senden(ADRESSE&0xFF);//Geraeteadresse senden
	 ir_byte_senden(n);//Code fuer Powertaste senden
	 ir_byte_senden(n^0xFF);//Inverser Code fuer Powertaste senden
	 ir_bit_senden(0);//abschliessendes Bit senden
	 do {milliwait(200); ir_repeat();}
	 while((PIND & (1<<INT0_BIT))==0);
	}
       else //Fehler: Reihe nicht gefunden
	{//test
	 for(uchar i=0;i<3;i++)  //test: LED 3 oder 4 mal blinken lassen
	  {
	   PORTB |= (1<<LED2_BIT); //LED2 einschalten
	   milliwait(200);
	   PORTB &= ~(1<<LED2_BIT); //LED2 aus
	   milliwait(200);
	  }
	}
       milliwait(50);
       taktzaehler=0;
       int0_flag=0;
       PORTC = 0; //PC alle Ausgaenge auf LOW (=aktiv)
      }

    }

   //Hier noch Sleep-Modus einschalten

  }

 return 0;
}
